"""Dataclasses for all structs in the Events contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name
# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import
# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return
from __future__ import annotations

from dataclasses import dataclass

from web3.types import ABIEvent, ABIEventParams

EventA = ABIEvent(
    anonymous=False,
    inputs=[
        ABIEventParams(indexed=True, name="who", type="address"),
        ABIEventParams(indexed=False, name="value", type="uint256"),
    ],
    name="EventA",
    type="event",
)

EventB = ABIEvent(
    anonymous=False,
    inputs=[],
    name="EventB",
    type="event",
)


@dataclass
class ErrorInfo:
    """Custom contract error information."""

    name: str
    selector: str
    signature: str
    inputs: list[ErrorParams]


@dataclass
class ErrorParams:
    """Parameter info for custom contract errors."""

    name: str
    solidity_type: str
    python_type: str



"""Utilities for working with hardhat."""

from typing import TypeGuard

from pypechain.hardhat.types import HardhatJson


def is_hardhat_json(val: object) -> TypeGuard[HardhatJson]:
    """Determines whether a json object is a HardhatJson."""
    required_keys = {"_format", "contractName", "sourceName"}

    has_required_keys = isinstance(val, dict) and required_keys.issubset(val.keys())
    if has_required_keys:
        _format: str = val.__dict__.get("_format", "")
        return _format == "hh-sol-artifact-1"
    return False


def get_bytecode_from_hardhat_json(json_abi: HardhatJson) -> str:
    """Gets the bytecode from a hardhat json file."""
    return json_abi.get("bytecode")



"""Tests recursive json discovery."""

import os

import pytest

from pypechain import pypechain_cli

# tests fixtures require passing the same name
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.fixture
def temp_dir_structure(tmpdir):
    """
    Creates a temporary directory structure with sample Solidity ABIs.
    """
    # Root directory
    root = tmpdir.mkdir("test_directory")

    # Subdirectories
    sub1 = root.mkdir("sub1")
    sub2 = root.mkdir("sub2")

    # Sample Solidity ABI JSONs
    abi1 = """
    {
        "abi":
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "name",
                    "outputs": [{"name": "", "type": "string"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
    }
    """

    abi2 = """
    {
        "abi":
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "symbol",
                    "outputs": [{"name": "", "type": "string"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
    }
    """

    abi3 = """
    {
        "abi":
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "decimals",
                    "outputs": [{"name": "", "type": "uint8"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
    }
    """

    # Write the ABIs to JSON files
    root.join("file1.json").write(abi1)
    sub1.join("file2.json").write(abi2)
    sub2.join("file3.json").write(abi3)

    return root


@pytest.mark.skip(reason="broken")
def test_recursive_json_discovery(temp_dir_structure):
    """
    Tests that the main() function recursively discovers JSON files and processes them.
    """
    # Output directory
    output_dir = temp_dir_structure.join("output")

    # Call main function
    pypechain_cli(argv=[str(temp_dir_structure), "--output-dir", str(output_dir)])

    # # Assertions
    assert os.path.exists(os.path.join(output_dir, "file1Contract.py"))
    assert os.path.exists(os.path.join(output_dir, "file1Types.py"))

    assert os.path.exists(os.path.join(output_dir, "file2Contract.py"))
    assert os.path.exists(os.path.join(output_dir, "file2Types.py"))

    assert os.path.exists(os.path.join(output_dir, "file3Contract.py"))
    assert os.path.exists(os.path.join(output_dir, "file3Types.py"))



"""Tests for overloading methods."""

from __future__ import annotations

import os

import pytest
from web3 import Web3
from web3.logs import WARN

from pypechain.test.events.types import EventsContract

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.mark.usefixtures("process_contracts")
class TestEvents:
    """Tests events emitted from the contracts."""

    @pytest.mark.skip()
    def test_process_receipt(self, w3: Web3):
        """Test that we can use event filters."""
        deployed_contract = EventsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        hash_a = deployed_contract.functions.emitOneEvent(0, "0x0000000000000000000000000000000000000000").transact()
        receipt_a = w3.eth.get_transaction_receipt(hash_a)
        hash_b = deployed_contract.functions.emitTwoEvents(0, "0x0000000000000000000000000000000000000000").transact()
        receipt_b = w3.eth.get_transaction_receipt(hash_b)

        deployed_contract.events.EventA.process_receipt(receipt_a, WARN)
        deployed_contract.events.EventB.process_receipt(receipt_b, WARN)

    def test_get_logs(self, w3):
        """Test that we can use event filters."""
        deployed_contract = EventsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        deployed_contract.functions.emitOneEvent(0, "0x0000000000000000000000000000000000000000").transact()
        deployed_contract.functions.emitTwoEvents(0, "0x0000000000000000000000000000000000000000").transact()

        event_a_logs = deployed_contract.events.EventA.get_logs(
            argument_filters={"who": "0x0000000000000000000000000000000000000000"},
            fromBlock=0,
        )
        event_b_logs = deployed_contract.events.EventB.get_logs(
            fromBlock=0,
        )
        assert len(list(event_a_logs)) == 2
        assert len(list(event_b_logs)) == 1

    def test_create_filter(self, w3):
        """Test that we can use event filters."""
        deployed_contract = EventsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        tx_hash = deployed_contract.functions.emitOneEvent(0, "0x0000000000000000000000000000000000000000").transact()
        w3.eth.wait_for_transaction_receipt(tx_hash)

        event_a_filter = deployed_contract.events.EventA.create_filter(
            argument_filters={"who": "0x0000000000000000000000000000000000000000"},
            fromBlock=0,
        )
        event_b_filter = deployed_contract.events.EventB.create_filter(
            fromBlock=0,
        )

        a_events = event_a_filter.get_new_entries()
        b_events = event_b_filter.get_new_entries()

        assert len(a_events) == 1
        assert len(b_events) == 0

        tx_hash = deployed_contract.functions.emitTwoEvents(0, "0x0000000000000000000000000000000000000000").transact()
        w3.eth.wait_for_transaction_receipt(tx_hash)

        a_events = event_a_filter.get_new_entries()
        b_events = event_b_filter.get_new_entries()

        assert len(a_events) == 1
        assert len(b_events) == 1



"""Types for Hardhat."""

from typing import Any, TypedDict

from web3.types import ABI


class HardhatJson(TypedDict):
    """Hardhat"""

    _format: str
    contractName: str
    sourceName: str
    abi: ABI
    bytecode: str
    deployedBytecode: str
    linkReferences: Any
    deployedLinkReferences: Any



"""Tests for overloading methods."""

from __future__ import annotations

import os

import pytest
from eth_account.account import Account
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.types import RPCEndpoint

from pypechain.test.deployment.types import ConstructorWithStructArgsContract
from pypechain.test.deployment.types.ConstructorWithStructArgsTypes import Config, Items

from .types import ConstructorNoArgsContract, ConstructorWithArgsContract, NoConstructorContract

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name
current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.fixture
def local_account(w3) -> LocalAccount:
    """Generates a random local account.

    Returns
    -------
    LocalAccount
        An account with a random private key and therefore public key.
    """
    local_account = Account.create("")

    amount_wei = int(100 * 1e18)
    params = [local_account.address, hex(amount_wei)]  # account, amount
    w3.provider.make_request(method=RPCEndpoint("anvil_setBalance"), params=params)

    return local_account


@pytest.mark.usefixtures("process_contracts")
class TestNoConstructorContract:
    """Tests deployment methods for contracts with no constructor"""

    def test_deployment_with_constructor_and_default_signer(self, w3):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        tx_hash = NoConstructorContract.factory(w3).constructor().transact()
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        assert receipt["status"] == 1

    def test_deployment_with_constructor_and_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        deployment_tx = NoConstructorContract.factory(w3).constructor().build_transaction()
        current_nonce = w3.eth.get_transaction_count(local_account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = local_account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        # Assert deployment success based on receipt
        assert receipt["status"] == 1  # or other relevant assertion

    def test_deploy_method_with_default_signer(self, w3, local_account):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        deployed_contract = NoConstructorContract.deploy(w3, local_account)
        name = deployed_contract.functions.name().call()
        assert name == "default"

    def test_deploy_method_with_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        deployed_contract = NoConstructorContract.deploy(w3, local_account)

        name = deployed_contract.functions.name().call()
        assert name == "default"


class TestConstructorNoArgs:
    """Tests deployment methods for contracts with a constructor that has no arguments."""

    def test_deployment_with_constructor_and_default_signer(self, w3):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        tx_hash = ConstructorNoArgsContract.factory(w3).constructor().transact()
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        assert receipt["status"] == 1

    def test_deployment_with_constructor_and_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        deployment_tx = ConstructorNoArgsContract.factory(w3).constructor().build_transaction()
        current_nonce = w3.eth.get_transaction_count(local_account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = local_account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        # Assert deployment success based on receipt
        assert receipt["status"] == 1  # or other relevant assertion

    def test_deploy_method_with_default_signer(self, w3, local_account):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        deployed_contract = ConstructorNoArgsContract.deploy(w3, local_account)
        name = deployed_contract.functions.name().call()
        assert name == "default"

    def test_deploy_method_with_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        deployed_contract = ConstructorNoArgsContract.deploy(w3, local_account)

        name = deployed_contract.functions.name().call()
        assert name == "default"


class TestConstructorWithArgs:
    """Tests deployment methods for contracts with a constructor that has arguments."""

    def test_deployment_with_constructor_and_default_signer(self, w3):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        tx_hash = ConstructorWithArgsContract.factory(w3).constructor("name").transact()
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        assert receipt["status"] == 1

    def test_deployment_with_constructor_and_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        deployment_tx = ConstructorWithArgsContract.factory(w3).constructor("name").build_transaction()
        current_nonce = w3.eth.get_transaction_count(local_account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = local_account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        # Assert deployment success based on receipt
        assert receipt["status"] == 1  # or other relevant assertion

    def test_deploy_method_with_default_signer(self, w3, local_account):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        deployed_contract = ConstructorWithArgsContract.deploy(
            w3, local_account, constructor_args=ConstructorWithArgsContract.ConstructorArgs("not default")
        )
        name = deployed_contract.functions.name().call()
        assert name == "not default"

    def test_deploy_method_with_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        deployed_contract = ConstructorWithArgsContract.deploy(
            w3, local_account, constructor_args=ConstructorWithArgsContract.ConstructorArgs("not default")
        )

        name = deployed_contract.functions.name().call()
        assert name == "not default"


class TestConstructorWithStructArgs:
    """Tests deployment methods for contracts with a constructor that has complex arguments."""

    def test_deployment_with_constructor_and_default_signer(self, w3):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        config = Config(name="name", items=Items(thing="thang", yesOrNo=True))
        tx_hash = ConstructorWithStructArgsContract.factory(w3).constructor(config).transact()
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        assert receipt["status"] == 1

    def test_deployment_with_constructor_and_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        config = Config(name="name", items=Items(thing="thang", yesOrNo=True))
        deployment_tx = ConstructorWithStructArgsContract.factory(w3).constructor(config).build_transaction()
        current_nonce = w3.eth.get_transaction_count(local_account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = local_account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        # Assert deployment success based on receipt
        assert receipt["status"] == 1  # or other relevant assertion

    def test_deploy_method_with_default_signer(self, w3, local_account):
        """Tests deployment using the constructor method instead of the deploy method."""
        # transact() uses the default signer in the web3 instance
        config = Config(name="name", items=Items(thing="thang", yesOrNo=True))
        args = ConstructorWithStructArgsContract.ConstructorArgs(config)
        deployed_contract = ConstructorWithStructArgsContract.deploy(w3, local_account, constructor_args=args)
        name = deployed_contract.functions.name().call()
        assert name == "name"

    def test_deploy_method_with_local_signer(self, w3: Web3, local_account: LocalAccount):
        """Tests deployment using the constructor method instead of the deploy method."""
        # Get the deployment transaction
        config = Config(name="name", items=Items(thing="thang", yesOrNo=True))
        args = ConstructorWithStructArgsContract.ConstructorArgs(config)
        deployed_contract = ConstructorWithStructArgsContract.deploy(w3, local_account, constructor_args=args)

        name = deployed_contract.functions.name().call()
        assert name == "name"




 
class OverloadedBalanceOfContractFunction(ContractFunction):
    """ContractFunction for the balanceOf method."""

    
    
    
    
    

    def __call__(self) -> OverloadedBalanceOfContractFunction: #type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> int:
        """returns int."""
        # Define the expected return types from the smart contract call
        
        return_types = int
        
        # Call the function
        
        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        return cast(int, rename_returned_types(structs, return_types, raw_values))
        
 


 
class OverloadedBalanceOfWhoContractFunction(ContractFunction):
    """ContractFunction for the balanceOfWho method."""

    
    
    
    
    

    def __call__(self, who: str) -> OverloadedBalanceOfWhoContractFunction: #type: ignore
        clone = super().__call__(dataclass_to_tuple(who))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> bool:
        """returns bool."""
        # Define the expected return types from the smart contract call
        
        return_types = bool
        
        # Call the function
        
        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        return cast(bool, rename_returned_types(structs, return_types, raw_values))
        
 



class OverloadedContractFunctions(ContractFunctions):
    """ContractFunctions for the Overloaded contract."""

    balanceOf: OverloadedBalanceOfContractFunction

    balanceOfWho: OverloadedBalanceOfWhoContractFunction

    def __init__(
         self,
         abi: ABI,
         w3: "Web3",
         address: ChecksumAddress | None = None,
         decode_tuples: bool | None = False,
     ) -> None:
         super().__init__(abi, w3, address, decode_tuples)
         self.balanceOf = OverloadedBalanceOfContractFunction.factory(
             "balanceOf",
             w3=w3,
             contract_abi=abi,
             address=address,
             decode_tuples=decode_tuples,
             function_identifier="balanceOf",
         )
         self.balanceOfWho = OverloadedBalanceOfWhoContractFunction.factory(
             "balanceOfWho",
             w3=w3,
             contract_abi=abi,
             address=address,
             decode_tuples=decode_tuples,
             function_identifier="balanceOfWho",
         )
         


"""A web3.py Contract class for the StructsA contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain"""

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# This file is bound to get very long depending on contract sizes.
# pylint: disable=too-many-lines

# methods are overriden with specific arguments instead of generic *args, **kwargs
# pylint: disable=arguments-differ

# consumers have too many opinions on line length
# pylint: disable=line-too-long


from __future__ import annotations

from typing import Any, Type, cast

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from typing_extensions import Self
from web3 import Web3
from web3.contract.contract import Contract, ContractConstructor, ContractFunction, ContractFunctions
from web3.exceptions import FallbackNotFound
from web3.types import ABI, BlockIdentifier, CallOverride, TxParams

from .IStructsTypes import InnerStruct, NestedStruct, SimpleStruct
from .StructsATypes import AStruct
from .utilities import rename_returned_types

structs = {
    "InnerStruct": InnerStruct,
    "NestedStruct": NestedStruct,
    "SimpleStruct": SimpleStruct,
    "AStruct": AStruct,
}


class StructsASingleNestedStructContractFunction(ContractFunction):
    """ContractFunction for the singleNestedStruct method."""

    def __call__(self) -> StructsASingleNestedStructContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> NestedStruct:
        """returns NestedStruct."""
        # Define the expected return types from the smart contract call

        return_types = NestedStruct

        # Call the function

        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        return cast(NestedStruct, rename_returned_types(structs, return_types, raw_values))


class StructsASingleSimpleStructContractFunction(ContractFunction):
    """ContractFunction for the singleSimpleStruct method."""

    def __call__(self) -> StructsASingleSimpleStructContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> SimpleStruct:
        """returns SimpleStruct."""
        # Define the expected return types from the smart contract call

        return_types = SimpleStruct

        # Call the function

        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        return cast(SimpleStruct, rename_returned_types(structs, return_types, raw_values))


class StructsAStructAContractFunction(ContractFunction):
    """ContractFunction for the structA method."""

    def __call__(self) -> StructsAStructAContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> AStruct:
        """returns AStruct."""
        # Define the expected return types from the smart contract call

        return_types = AStruct

        # Call the function

        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        return cast(AStruct, rename_returned_types(structs, return_types, raw_values))


class StructsAContractFunctions(ContractFunctions):
    """ContractFunctions for the StructsA contract."""

    singleNestedStruct: StructsASingleNestedStructContractFunction

    singleSimpleStruct: StructsASingleSimpleStructContractFunction

    structA: StructsAStructAContractFunction

    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
        decode_tuples: bool | None = False,
    ) -> None:
        super().__init__(abi, w3, address, decode_tuples)
        self.singleNestedStruct = StructsASingleNestedStructContractFunction.factory(
            "singleNestedStruct",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="singleNestedStruct",
        )
        self.singleSimpleStruct = StructsASingleSimpleStructContractFunction.factory(
            "singleSimpleStruct",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="singleSimpleStruct",
        )
        self.structA = StructsAStructAContractFunction.factory(
            "structA",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="structA",
        )


structsa_abi: ABI = cast(
    ABI,
    [
        {
            "type": "function",
            "name": "singleNestedStruct",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct IStructs.NestedStruct",
                    "components": [
                        {"name": "intVal", "type": "uint256", "internalType": "uint256"},
                        {"name": "strVal", "type": "string", "internalType": "string"},
                        {
                            "name": "innerStruct",
                            "type": "tuple",
                            "internalType": "struct IStructs.InnerStruct",
                            "components": [{"name": "boolVal", "type": "bool", "internalType": "bool"}],
                        },
                    ],
                }
            ],
            "stateMutability": "pure",
        },
        {
            "type": "function",
            "name": "singleSimpleStruct",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct IStructs.SimpleStruct",
                    "components": [
                        {"name": "intVal", "type": "uint256", "internalType": "uint256"},
                        {"name": "strVal", "type": "string", "internalType": "string"},
                    ],
                }
            ],
            "stateMutability": "pure",
        },
        {
            "type": "function",
            "name": "structA",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct StructsA.AStruct",
                    "components": [
                        {"name": "intVal", "type": "uint256", "internalType": "uint256"},
                        {"name": "strVal", "type": "string", "internalType": "string"},
                    ],
                }
            ],
            "stateMutability": "pure",
        },
    ],
)


class StructsAContract(Contract):
    """A web3.py Contract class for the StructsA contract."""

    abi: ABI = structsa_abi
    # We change `bytecode` as needed for linking, but keep
    # `_raw_bytecode` unchanged as an original copy.
    # pylint: disable=line-too-long
    _raw_bytecode: HexStr | None = HexStr(
        "0x608060405234801561001057600080fd5b5061024e806100206000396000f3fe608060405234801561001057600080fd5b50600436106100415760003560e01c806311ba731f146100465780636456bc7914610046578063c567c2f3146100a5575b600080fd5b604080518082018252600081526060602091820152815180830183526001815282518084018452601081526f596f7520617265206e756d626572203160801b8184015291810191909152905161009c91906101bf565b60405180910390f35b6100ad6100ba565b60405161009c91906101d9565b6100c261011c565b6040518060600160405280600181526020016040518060400160405280601081526020016f596f7520617265206e756d626572203160801b8152508152602001604051806020016040528060011515815250815250905090565b6040518060600160405280600081526020016060815260200161014d60405180602001604052806000151581525090565b905290565b6000815180845260005b818110156101785760208185018101518683018201520161015c565b506000602082860101526020601f19601f83011685010191505092915050565b8051825260006020820151604060208501526101b76040850182610152565b949350505050565b6020815260006101d26020830184610198565b9392505050565b602081528151602082015260006020830151606060408401526101ff6080840182610152565b905060408401515115156060840152809150509291505056fea2646970667358221220bc1b4d24faa73bb69c7c733cb49b490e847da06a9083c0d28e077d44105cf11564736f6c63430008160033"
    )

    def __init__(self, address: ChecksumAddress | None = None) -> None:
        try:
            # Initialize parent Contract class
            super().__init__(address=address)
            self.functions = StructsAContractFunctions(structsa_abi, self.w3, address)  # type: ignore

        except FallbackNotFound:
            print("Fallback function not found. Continuing...")

    functions: StructsAContractFunctions

    @classmethod
    def constructor(cls) -> ContractConstructor:  # type: ignore
        """Creates a transaction with the contract's constructor function.

        Parameters
        ----------

        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.

        """
        cls.bytecode = cls._raw_bytecode
        if cls.bytecode is not None:

            # bytecode needs to be in hex for web3
            cls.bytecode = HexBytes(cls.bytecode)

        return super().constructor()

    @classmethod
    def deploy(cls, w3: Web3, account: LocalAccount | ChecksumAddress) -> Self:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        deployer = cls.factory(w3=w3)
        constructor_fn = deployer.constructor()

        # if an address is supplied, try to use a web3 default account
        if isinstance(account, str):
            tx_hash = constructor_fn.transact({"from": account})
            tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

            deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
            return deployed_contract

        # otherwise use the account provided.
        deployment_tx = constructor_fn.build_transaction()
        current_nonce = w3.eth.get_transaction_count(account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
        return deployed_contract

    @classmethod
    def factory(cls, w3: Web3, class_name: str | None = None, **kwargs: Any) -> Type[Self]:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        class_name: str | None
            The instance class name.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        contract = super().factory(w3, class_name, **kwargs)
        contract.functions = StructsAContractFunctions(structsa_abi, w3, None)

        return contract



"""Conversion helpers."""

from __future__ import annotations

from dataclasses import fields, is_dataclass
from typing import Any, Tuple, TypeVar, cast

T = TypeVar("T")


def tuple_to_dataclass(cls: type[T], tuple_data: Any | Tuple[Any, ...]) -> T:
    """
    Converts a tuple (including nested tuples) to a dataclass instance.  If cls is not a dataclass,
    then the data will just be passed through this function.

    Parameters
    ----------
    cls: type[T]
        The dataclass type to which the tuple data is to be converted.
    tuple_data: Any | Tuple[Any, ...]
        A tuple (or nested tuple) of values to convert into a dataclass instance.

    Returns
    -------
    T
        Either an instance of cls populated with data from tuple_data or tuple_data itself.
    """
    if not is_dataclass(cls):
        return cast(T, tuple_data)

    field_types = {field.name: field.type for field in fields(cls)}
    field_values = {}

    for (field_name, field_type), value in zip(field_types.items(), tuple_data):
        if is_dataclass(field_type):
            # Recursively convert nested tuples to nested dataclasses
            field_values[field_name] = tuple_to_dataclass(field_type, value)
        elif isinstance(value, tuple) and not getattr(field_type, "_name", None) == "Tuple":
            # If it's a tuple and the field is not intended to be a tuple, assume it's a nested dataclass
            field_values[field_name] = tuple_to_dataclass(field_type, value)
        else:
            # Otherwise, set the primitive value directly
            field_values[field_name] = value

    return cls(**field_values)



"""A web3.py Contract class for the NoConstructor contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain"""

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# This file is bound to get very long depending on contract sizes.
# pylint: disable=too-many-lines

# methods are overriden with specific arguments instead of generic *args, **kwargs
# pylint: disable=arguments-differ

# consumers have too many opinions on line length
# pylint: disable=line-too-long


from __future__ import annotations

from typing import Any, Type, cast

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from typing_extensions import Self
from web3 import Web3
from web3.contract.contract import Contract, ContractConstructor, ContractFunction, ContractFunctions
from web3.exceptions import FallbackNotFound
from web3.types import ABI, BlockIdentifier, CallOverride, TxParams

from .utilities import dataclass_to_tuple, rename_returned_types

structs = {}


class NoConstructorNameContractFunction(ContractFunction):
    """ContractFunction for the name method."""

    def __call__(self) -> NoConstructorNameContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> str:
        """returns str."""
        # Define the expected return types from the smart contract call

        return_types = str

        # Call the function

        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        return cast(str, rename_returned_types(structs, return_types, raw_values))


class NoConstructorSetNameContractFunction(ContractFunction):
    """ContractFunction for the setName method."""

    def __call__(self, name: str) -> NoConstructorSetNameContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(name))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class NoConstructorContractFunctions(ContractFunctions):
    """ContractFunctions for the NoConstructor contract."""

    name: NoConstructorNameContractFunction

    setName: NoConstructorSetNameContractFunction

    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
        decode_tuples: bool | None = False,
    ) -> None:
        super().__init__(abi, w3, address, decode_tuples)
        self.name = NoConstructorNameContractFunction.factory(
            "name",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="name",
        )
        self.setName = NoConstructorSetNameContractFunction.factory(
            "setName",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="setName",
        )


noconstructor_abi: ABI = cast(
    ABI,
    [
        {
            "type": "function",
            "name": "name",
            "inputs": [],
            "outputs": [{"name": "", "type": "string", "internalType": "string"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "setName",
            "inputs": [{"name": "_name", "type": "string", "internalType": "string"}],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
    ],
)


class NoConstructorContract(Contract):
    """A web3.py Contract class for the NoConstructor contract."""

    abi: ABI = noconstructor_abi
    # We change `bytecode` as needed for linking, but keep
    # `_raw_bytecode` unchanged as an original copy.
    # pylint: disable=line-too-long
    _raw_bytecode: HexStr | None = HexStr(
        "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"
    )

    def __init__(self, address: ChecksumAddress | None = None) -> None:
        try:
            # Initialize parent Contract class
            super().__init__(address=address)
            self.functions = NoConstructorContractFunctions(noconstructor_abi, self.w3, address)  # type: ignore

        except FallbackNotFound:
            print("Fallback function not found. Continuing...")

    functions: NoConstructorContractFunctions

    @classmethod
    def constructor(cls) -> ContractConstructor:  # type: ignore
        """Creates a transaction with the contract's constructor function.

        Parameters
        ----------

        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.

        """
        cls.bytecode = cls._raw_bytecode
        if cls.bytecode is not None:

            # bytecode needs to be in hex for web3
            cls.bytecode = HexBytes(cls.bytecode)

        return super().constructor()

    @classmethod
    def deploy(cls, w3: Web3, account: LocalAccount | ChecksumAddress) -> Self:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        deployer = cls.factory(w3=w3)
        constructor_fn = deployer.constructor()

        # if an address is supplied, try to use a web3 default account
        if isinstance(account, str):
            tx_hash = constructor_fn.transact({"from": account})
            tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

            deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
            return deployed_contract

        # otherwise use the account provided.
        deployment_tx = constructor_fn.build_transaction()
        current_nonce = w3.eth.get_transaction_count(account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
        return deployed_contract

    @classmethod
    def factory(cls, w3: Web3, class_name: str | None = None, **kwargs: Any) -> Type[Self]:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        class_name: str | None
            The instance class name.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        contract = super().factory(w3, class_name, **kwargs)
        contract.functions = NoConstructorContractFunctions(noconstructor_abi, w3, None)

        return contract



"""Dataclasses for all structs in the Example contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name
# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import
# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return
from __future__ import annotations

from dataclasses import dataclass

from web3.types import ABIEvent, ABIEventParams


@dataclass
class SimpleStruct:
    """SimpleStruct struct."""

    intVal: int
    strVal: str


@dataclass
class InnerStruct:
    """InnerStruct struct."""

    boolVal: bool


@dataclass
class NestedStruct:
    """NestedStruct struct."""

    intVal: int
    strVal: str
    innerStruct: InnerStruct


Flip = ABIEvent(
    anonymous=False,
    inputs=[
        ABIEventParams(indexed=False, name="flip", type="uint256"),
    ],
    name="Flip",
    type="event",
)

Flop = ABIEvent(
    anonymous=False,
    inputs=[
        ABIEventParams(indexed=False, name="flop", type="uint256"),
    ],
    name="Flop",
    type="event",
)


@dataclass
class ErrorInfo:
    """Custom contract error information."""

    name: str
    selector: str
    signature: str
    inputs: list[ErrorParams]


@dataclass
class ErrorParams:
    """Parameter info for custom contract errors."""

    name: str
    solidity_type: str
    python_type: str


WrongChoiceError = ErrorInfo(
    inputs=[
        ErrorParams(name="answer", python_type="int", solidity_type="uint8"),
        ErrorParams(name="errorMessage", python_type="str", solidity_type="string"),
    ],
    name="WrongChoice",
    selector="0xc13b30d4",
    signature="WrongChoice(uint8,string)",
)



"""Sorting utilities."""


def get_intersection_and_unique(lists: list[list[str]]) -> tuple[set[str], set[str]]:
    """Process a list of lists of strings to get the intersection and unique values.

    The intersection is a set of strings that occur in all sub-lists.
    The unique values are strings that only occur in one sub-list.

    Parameters
    ----------
    lists : list[list[str]]
        A list of lists of strings, where each sub-list is an entity to compute sets over.

    Returns
    -------
    tuple[set[str], set[str]]
        The (intersection, unique_values) sets
    """
    intersection = set(lists[0])
    for lst in lists[1:]:
        intersection &= set(lst)
    string_counts = {}
    for lst in lists:
        for item in set(lst):
            string_counts[item] = string_counts.get(item, 0) + 1
    unique_values = {item for item, count in string_counts.items() if count == 1}
    return (intersection, unique_values)



"""Utilities for solidity contract ABIs."""

from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, NamedTuple, Sequence, TypeGuard, cast

from eth_utils.abi import collapse_if_tuple, function_signature_to_4byte_selector
from eth_utils.hexadecimal import encode_hex
from web3.types import ABI, ABIElement, ABIError, ABIEvent, ABIFunction, ABIFunctionComponents, ABIFunctionParams

from pypechain.foundry.types import FoundryJson
from pypechain.hardhat.types import HardhatJson
from pypechain.hardhat.utilities import is_hardhat_json
from pypechain.solc.types import SolcJson
from pypechain.solc.utilities import is_solc_json
from pypechain.utilities.bytecode import get_bytecode_from_json, get_bytecode_link_references_from_json, is_foundry_json
from pypechain.utilities.format import avoid_python_keywords, capitalize_first_letter_only
from pypechain.utilities.types import LinkReferences, solidity_to_python_type

# These are exactly the same sowe just rename.
ABIErrorParams = ABIFunctionParams


class AbiJson(NamedTuple):
    """A JSON representation of a solidity contract's Application Boundary Interface."""

    abi: Sequence[ABIElement]


def load_abi(abi_path: str) -> AbiJson:
    """Loads the abi file into a json.

    Parameters
    ----------
    abi_path : str
        Where the abi json is location.

    Returns
    -------
    AbiJson
        A named tuple representation of an abi json file.
    """

    with open(abi_path, "r", encoding="utf-8") as abi_file:
        data = json.load(abi_file)

        # Assuming that the ABI data structure is at the top level of the JSON
        # (i.e., the file is a list of ABI items):
        abi_items = [cast(ABIElement, item) for item in data]

        return AbiJson(abi=abi_items)


def is_abi_function(item: ABIElement) -> TypeGuard[ABIFunction]:
    """Typeguard function for ABIFunction.

    Parameters
    ----------
    item:  Any
        The item we are confirming is an ABIFunction

    Returns
    -------
    TypeGuard[ABIFunction]
    """
    # Check if the required keys exist
    required_keys = ["type", "inputs"]

    # Check if the required keys exist
    if not all(key in item for key in required_keys):
        return False

    # Check if the type is a function
    if item.get("type") not in ["function", "constructor", "fallback", "receive"]:
        return False

    return True


def get_abi_constructor(abi: ABI) -> ABIFunction | None:
    """Returns the constructor item if it exists.

    Parameters
    ----------
    abi: ABI
        The contract's abi json to parse.

    Returns
    -------
    ABIFunction | None
        The constructor information returned as an abi function since they have the same pattern.
    """

    result = next((x for x in abi if x.get("type", "") == "constructor"), None)
    return result if result and is_abi_function(result) else None


def is_abi_constructor(item: ABIElement) -> TypeGuard[ABIFunction]:
    """Typeguard function for ABIFunction.

    Parameters
    ----------
    item:  Any
        The item we are confirming is an ABIFunction

    Returns
    -------
    TypeGuard[ABIFunction]
    """
    # Check if the required keys exist
    required_keys = ["type", "inputs"]

    # Check if the required keys exist
    if not all(key in item for key in required_keys):
        return False

    # Check if the type is a constructor function
    if item.get("type") != "constructor":
        return False

    return True


def is_abi_event(item: ABIElement) -> TypeGuard[ABIEvent]:
    """Typeguard function for ABIEvent.

    Parameters
    ----------
    item:  Any
        The item we are confirming is an ABIFunction

    Returns
    -------
    TypeGuard[ABIEvent]
    """
    # Check if the required keys exist
    required_keys = ["type", "name", "inputs"]

    # Check if the required keys exist
    if not all(key in item for key in required_keys):
        return False

    # Check if the type is "event"
    if item.get("type") != "event":
        return False

    return True


def is_abi_error(item: ABIElement) -> TypeGuard[ABIError]:
    """Typeguard function for ABIError.

    Parameters
    ----------
    item:  Any
        The item we are confirming is an ABIError

    Returns
    -------
    TypeGuard[ABIError]
    """
    # Check if the required keys exist
    required_keys = ["type", "name", "inputs"]

    # Check if the required keys exist
    if not all(key in item for key in required_keys):
        return False

    # Check if the type is "event"
    if item.get("type") != "error":
        return False

    return True


def filter_abi_items_by_type(item_type: str | list[str], contract_abi: ABI) -> list[ABIElement]:
    """Filters ABIItems by type.

    Parameters
    ----------
    item_type : str | list[str]
        Item or list of items to filter by
    contract_abi : ABI
        The ABI to filter

    Returns
    -------
    list[ABIFunction | ABIEvent]

    """

    if isinstance(item_type, str):
        item_type = [item_type]

    return [abi for abi in contract_abi if abi.get("type", "") in item_type]


@dataclass
class StructInfo:
    """Solidity struct information needed for codegen."""

    name: str
    contract_name: str
    values: list[StructValue]


@dataclass
class StructValue:
    """The name and python type of a solidity struct value."""

    name: str
    # TODO: type this better with an exahaustive list of python type strings.  Even better, maybe a
    # mapping from solidity to python types?
    solidity_type: str
    python_type: str
    is_struct: bool
    contract_name: str | None


@dataclass
class EventInfo:
    """Solidity struct information needed for codegen."""

    name: str
    anonymous: bool
    inputs: list[EventParams]


@dataclass
class EventParams:
    """Solidity struct information needed for codegen."""

    indexed: bool
    name: str
    solidity_type: str
    python_type: str


@dataclass
class ErrorInfo:
    """Solidity struct information needed for codegen."""

    name: str
    selector: str
    signature: str
    inputs: list[ErrorParams]


@dataclass
class ErrorParams:
    """Solidity struct information needed for codegen."""

    name: str
    solidity_type: str
    python_type: str


def get_structs(
    function_params: Sequence[ABIFunctionParams] | Sequence[ABIFunctionComponents],
    structs: dict[str, StructInfo] | None = None,
) -> dict[str, StructInfo]:
    """Recursively gets all the structs for a contract by walking all function parameters.

     Pseudo code of the shape of a Sequence[ABIFunctionParams]:
     [
        { #ABIFunctionParams
            name:
            type: asdf
            internalType:
            components:
        },
        { #ABIFunctionParams
            name:
            type:
            internalType:
            components: [ #Sequence[ABIFunctionComponents]
                { #ABIFunctionComponents
                    name:
                    type:
                },
                { #ABIFunctionComponents
                    name:
                    type:
                    internalType:
                    components: # Sequence[ABIFunctionComponents]
                }
            ]
        },
    ]

    Parameters
    ----------
    function_params : Sequence[ABIFunctionParams] | Sequence[ABIFunctionComponents]
        The function parameter lists to parse structs from.

    structs : dict[str, StructInfo] | None = None
         A dictionary of StructInfos keyed by struct name that is populated during recursion.

    Returns
    -------
    dict[str, StructInfo]
        The completed dictionary of StructInfos keyed by struct name.
    """
    if structs is None:
        structs = {}
    for param in function_params:
        components = param.get("components")
        internal_type = cast(str, param.get("internalType", ""))

        # If we find a struct, we'll add it to the dict of StructInfo's
        if is_struct(internal_type) and components:
            struct_name = get_struct_name(param)
            struct_file_name = get_struct_contract_name(param)
            struct_values: list[StructValue] = []

            # Walk over the components of the struct
            for component in components:
                component_internal_type = cast(str, component.get("internalType", ""))

                # Do recursion if nested struct
                contract_name: str | None = None
                if is_struct(component_internal_type):
                    structs = get_structs([component], structs)
                    contract_name = get_struct_contract_name(component)

                # Get the component name and type
                component_name = get_param_name(component)  # strip [] if it is an array
                component_type = (
                    get_struct_type(component) if is_struct(component_internal_type) else component.get("type", "")
                )
                python_type = solidity_to_python_type(
                    component_type, custom_types=[struct.name for struct in structs.values()]
                )

                # Collect information
                struct_values.append(
                    StructValue(
                        name=component_name,
                        solidity_type=component_type,
                        python_type=python_type,
                        is_struct=is_struct(component_internal_type),
                        contract_name=contract_name,
                    )
                )

            # Add the struct to the dict
            structs[f"{struct_file_name}.{struct_name}"] = StructInfo(
                name=struct_name, contract_name=struct_file_name, values=struct_values
            )

    return structs


def get_structs_for_abi(abi: ABI) -> list[StructInfo]:
    """Gets all the structs for a given abi.
    These are found by parsing function inputs and outputs for internalType's.

    Parameters
    ----------
    abi : ABI
        An Application Boundary Interface object.

    Returns
    -------
    dict[str, StructInfo]
        A dictionary of StructInfos keyed by name.
    """
    structs: dict[str, StructInfo] = {}
    for item in abi:
        if is_abi_function(item):
            fn_inputs = item.get("inputs")
            fn_outputs = item.get("outputs")
            if fn_inputs:
                input_structs = get_structs(fn_inputs, structs)
                structs.update(input_structs)
            if fn_outputs:
                output_structs = get_structs(fn_outputs, structs)
                structs.update(output_structs)
    return list(structs.values())


def is_struct(internal_type: str) -> bool:
    """Returns True if the internal type of the parameter is a solidity struct.

    Parameters
    ----------
    internal_type : str
        The internalType attribute of an ABIFunctionParams or ABIFunctionComponents.  If the
        internal type has the form 'struct ContractName.StructName' then we know we are dealing with
        solidity struct.  Otherwise it will be equivalent to the 'type' attribute.

    Returns
    -------
    bool
        If the type is a struct.
    """
    # internal_type looks like 'struct ContractName.StructName' if they are structs
    return bool(internal_type.startswith("struct"))


def get_events_for_abi(abi: ABI) -> list[EventInfo]:
    """Gets all the events for a given abi.

    Parameters
    ----------
    abi : ABI
        An Application Boundary Interface object.

    Returns
    -------
    list[ABIEvent]
        A dictionary of StructInfos keyed by name.
    """

    events: list[EventInfo] = []

    anonymous_event_counter: int = 0

    for item in abi:
        if is_abi_event(item):
            event: ABIEvent = item
            event_inputs = event.get("inputs", [])
            inputs: list[EventParams] = []

            for i in event_inputs:
                indexed = i.get("indexed", False)
                name = i.get("name", "annonymous")
                solidity_type = i.get("type")
                if not solidity_type:
                    raise ValueError("Type not known for event input.")

                python_type = solidity_to_python_type(solidity_type)
                event_input = EventParams(
                    indexed=indexed,
                    name=name,
                    solidity_type=solidity_type,
                    python_type=python_type,
                )

                inputs.append(event_input)

            anonymous = item.get("anonymous", False)
            # TODO add test for multiple anonymous events
            events.append(
                EventInfo(
                    name=item.get("name", f"Annonymous{anonymous_event_counter or None}"),
                    anonymous=anonymous,
                    inputs=inputs,
                )
            )

    return events


def get_errors_for_abi(abi: ABI) -> list[ErrorInfo]:
    """Gets all the errors for a given abi.

    Parameters
    ----------
    abi : ABI
        An Application Boundary Interface object.

    Returns
    -------
    list[ErrorInfo]
        A dictionary of StructInfos keyed by name.
    """

    errors: list[ErrorInfo] = []

    for item in abi:
        if is_abi_error(item):
            error_inputs = item.get("inputs", [])
            inputs: list[ErrorParams] = []

            for i in error_inputs:
                name = i.get("name", "")
                solidity_type = i.get("type")
                if not solidity_type:
                    raise ValueError("Type not known for event input.")

                python_type = solidity_to_python_type(solidity_type)
                error_input = ErrorParams(
                    name=name,
                    solidity_type=solidity_type,
                    python_type=python_type,
                )

                inputs.append(error_input)

            signature = _abi_to_signature(dict(item)).replace(" ", "")
            selector = encode_hex(function_signature_to_4byte_selector(signature))

            errors.append(
                ErrorInfo(
                    name=item.get("name"),
                    selector=selector,
                    signature=signature,
                    inputs=inputs,
                )
            )

    return errors


def _abi_to_signature(abi: dict[str, Any]) -> str:
    fn_input_types = ",".join([collapse_if_tuple(abi_input) for abi_input in abi.get("inputs", [])])
    function_signature = f"{abi['name']}({fn_input_types})"
    return function_signature


def get_struct_name(
    param_or_component: ABIFunctionParams | ABIFunctionComponents,
) -> str:
    """Returns the name of the given struct.

    For example, a struct in an abi json will look like:

    {
    "components": [
      # components listed here
    ],
    "internalType": "struct ContractName.SimpleStruct",
    "name": "simpleStruct",
    "type": "tuple"
    }

    We are assuming 'internalType's value to have the form:
    'struct [ContractName].[StructName]' or 'struct [StructName]'

    Parameters
    ----------
    param_or_component: ABIFunctionParams | ABIFunctionComponents


    Returns
    -------
    str
        The name of the item.
    """
    internal_type = cast(str, param_or_component.get("internalType", ""))
    # grab subtype if there is one
    if "." in internal_type:
        internal_type = internal_type.split(".")[1]
    # it is possible that the internal type has a "struct" label
    # we want to strip that to only include the struct name itself
    internal_type = internal_type.replace("struct ", "")
    struct_name = internal_type.rstrip("[]")
    return capitalize_first_letter_only(struct_name)


def get_struct_type(
    param_or_component: ABIFunctionParams | ABIFunctionComponents,
) -> str:
    """Returns the type of the given struct.

    This works the same as `get_struct_name`, except that we account for arrays by
    wrapping the name in `list[...]`.

    Parameters
    ----------
    param_or_component: ABIFunctionParams | ABIFunctionComponents

    Returns
    -------
    str
        The type of the item.
    """
    internal_type = cast(str, param_or_component.get("internalType", ""))
    # grab subtype if there is one
    if "." in internal_type:
        internal_type = internal_type.split(".")[1]
    # it is possible that the internal type has a "struct" label
    # we want to strip that to only include the struct name itself
    internal_type = internal_type.replace("struct ", "")
    if internal_type[-2:] == "[]":  # ends with [] indicates an array
        return "list[" + get_struct_name(param_or_component) + "]"
    return get_struct_name(param_or_component)


def get_struct_contract_name(
    param_or_component: ABIFunctionParams | ABIFunctionComponents,
) -> str:
    """Returns the contract name that the given struct is defined in.

    For example, a struct in an abi json will look like:

    {
    "components": [
      # components listed here
    ],
    "internalType": "struct ContractName.SimpleStruct",
    "name": "simpleStruct",
    "type": "tuple"
    }

    We are assuming 'internalType's value to have the form:
    'struct [ContractName].[StructName]'

    Parameters
    ----------
    param : ABIFunctionParams | ABIFunctionComponents


    Returns
    -------
    str
        The name of the item.
    """
    internal_type = cast(str, param_or_component.get("internalType", ""))
    contract_name_and_struct_name = internal_type.replace("struct ", "")
    contract_name = contract_name_and_struct_name.split(".")[0]
    return capitalize_first_letter_only(contract_name)


def get_param_name(
    param_or_component: ABIFunctionParams | ABIFunctionComponents,
) -> str:
    """Returns the name for a given ABIFunctionParams or ABIFunctionComponents.

    If the item is a struct, then we pull the name from the internalType attribute, otherwise we use
    the name if available.

    Parameters
    ----------
    param : ABIFunctionParams | ABIFunctionComponents

    Returns
    -------
    str
        The name of the item.
    """
    return param_or_component.get("name", "").lstrip("_").rstrip("[]")


@dataclass
class AbiInfo:
    """Information about an ABI"""

    abi: ABI
    bytecode: str
    contract_name: str
    # Bytecode link references is a dictionary keyed by a contract name,
    # valued with the hex code placeholder in the bytecode.
    bytecode_link_references: list[LinkReferences]


def load_abi_infos_from_file(file_path: Path) -> list[AbiInfo]:
    """Loads a contract ABI from a file.

    Parameters
    ----------
    file_path : Path
        The path to the ABI file.

    Returns
    -------
    tuple[ABI, str]
        ABI: An object containing the contract's abi.
        str: The bytecode.
    """

    with open(file_path, "r", encoding="utf-8") as file:
        json_file = json.load(file)

        # solc jsons list all the contracts and their abis in one file, so we process it first.
        if is_solc_json(json_file):
            return _get_abis_from_solc_json(json_file)

        if is_hardhat_json(json_file):
            abi = get_abi_from_json(json_file)
            bytecode = get_bytecode_from_json(json_file)
            # hardhat jsons have the contract name as a field.
            contract_name = json_file.get("contractName")
            # TODO get bytecode link references and add here
            return [AbiInfo(abi=abi, bytecode=bytecode, contract_name=contract_name, bytecode_link_references=[])]

        if is_foundry_json(json_file):
            abi = get_abi_from_json(json_file)
            bytecode = get_bytecode_from_json(json_file)
            bytecode_link_references = get_bytecode_link_references_from_json(json_file)
            # foundry saves contracts to self-named files.
            contract_name = file_path.name.removesuffix(".json")
            return [
                AbiInfo(
                    abi=abi,
                    bytecode=bytecode,
                    contract_name=contract_name,
                    bytecode_link_references=bytecode_link_references,
                )
            ]

        raise ValueError("Unknown ABI Json format")


def get_abi_items(abi: ABI) -> list[ABIElement]:
    """Gets all of the functions and events in the ABI.

    Parameters
    ----------
    file_path : Path
        the file path to the ABI.

    Returns
    -------
    List[ABIElement]
        A list of all the ABI elements.
    """

    abi_functions_and_events = filter_abi_items_by_type(["function", "event"], abi)
    return abi_functions_and_events


def get_function_parameter_names(
    parameters: Sequence[ABIFunctionParams | ABIFunctionComponents],
) -> list[str]:
    """Parses a list of ABIFunctionParams or ABIFUnctionComponents and returns a list of parameter
    names."""

    stringified_function_parameters: list[str] = []
    arg_counter: int = 1
    for _input in parameters:
        if name := get_param_name(_input):
            stringified_function_parameters.append(avoid_python_keywords(name))
        else:
            name = f"arg{arg_counter}"
            stringified_function_parameters.append(avoid_python_keywords(name))
        arg_counter += 1
    return stringified_function_parameters


def get_input_names(function: ABIFunction) -> list[str]:
    """Returns function input name strings for jinja templating.

    i.e. for the solidity function signature:
    function doThing(address who, uint256 amount, bool flag, bytes extraData)

    the following list would be returned:
    ['who', 'amount', 'flag', 'extraData']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.

    Returns
    -------
    list[str]
        A list of function names i.e. ['arg1', 'arg2']
    """
    return get_function_parameter_names(function.get("inputs", []))


def get_output_names(function: ABIFunction) -> list[str]:
    """Returns function output name strings for jinja templating.

    i.e. for the solidity function signature: function doThing() returns (address who, uint256
    amount, bool flag, bytes extraData)

    the following list would be returned: ['who', 'amount', 'flag', 'extraData']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.

    Returns
    -------
    list[str]
        A list of function names i.e. [{name: 'arg1', type: 'int'}, { name: 'TransferInfo',
        components: [{
            name: 'from', type: 'str'}, name: '
        }]]
    """
    return get_function_parameter_names(function.get("outputs", []))


def get_input_names_and_types(function: ABIFunction) -> list[str]:
    """Returns function input name/type strings for jinja templating.

    i.e. for the solidity function signature: function doThing(address who, uint256 amount, bool
    flag, bytes extraData)

    the following list would be returned: ['who: str', 'amount: int', 'flag: bool', 'extraData:
    bytes']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.

    Returns
    -------
    list[str]
        A list of function names and corresponding python types, i.e. ['arg1: str', 'arg2: bool']
    """
    return _get_names_and_types(function, "inputs")


def get_input_types(function: ABIFunction) -> list[str]:
    """Returns function input type strings for jinja templating.

    i.e. for the solidity function signature: function doThing(address who, uint256 amount, bool
    flag, bytes extraData)

    the following list would be returned: ['str', 'int', 'bool', 'bytes']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.

    Returns
    -------
    list[str]
        A list of function python types, i.e. ['str', 'bool']
    """
    return _get_param_types(function, "inputs")


def get_output_types(function: ABIFunction) -> list[str]:
    """Returns function output type strings for jinja templating.

    i.e. for the solidity function signature: function doThing(address who, uint256 amount, bool
    flag, bytes extraData)

    the following list would be returned: ['str', 'int', 'bool', 'bytes']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.

    Returns
    -------
    list[str]
        A list of function python types, i.e. ['str', 'bool']
    """
    return _get_param_types(function, "outputs")


def get_output_names_and_types(function: ABIFunction) -> list[str]:
    """Returns function input name/type strings for jinja templating.

    i.e. for the solidity function signature: function doThing(address who, uint256 amount, bool
    flag, bytes extraData)

    the following list would be returned: ['who: str', 'amount: int', 'flag: bool', 'extraData:
    bytes']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.

    Returns
    -------
    list[str]
        A list of function names and corresponding python types, i.e. ['arg1: str', 'arg2: bool']
    """
    return _get_names_and_types(function, "outputs")


def _get_names_and_types(function: ABIFunction, parameters_type: Literal["inputs", "outputs"]) -> list[str]:
    """Returns function input or output name/type strings for jinja templating.

    i.e. for the solidity function signature: function doThing(address who, uint256 amount, bool
    flag, bytes extraData)

    the following list would be returned: ['who: str', 'amount: int', 'flag: bool', 'extraData:
    bytes']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.
    parameters_type : Literal["inputs", "outputs"]
        If we are looking at the inputs or outputs of a function.

    Returns
    -------
    list[str]
        A list of function names and corresponding python types, i.e. ['arg1: str', 'arg2: bool']
    """
    stringified_function_parameters: list[str] = []
    for index, param in enumerate(function.get(parameters_type, []), start=1):
        name = get_param_name(param)
        if not name:
            name = f"arg{index}"
        python_type = get_param_type(param)
        stringified_function_parameters.append(f"{avoid_python_keywords(name)}: {python_type}")
    return stringified_function_parameters


def _get_param_types(function: ABIFunction, parameters_type: Literal["inputs", "outputs"]) -> list[str]:
    """Returns function input or output type strings for jinja templating.

    i.e. for the solidity function signature: function doThing(address who, uint256 amount, bool
    flag, bytes extraData)

    the following list would be returned: ['who: str', 'amount: int', 'flag: bool', 'extraData:
    bytes']

    Parameters
    ----------
    function : ABIFunction
        A web3 dict of an ABI function description.
    parameters_type : Literal["inputs", "outputs"]
        If we are looking at the inputs or outputs of a function.

    Returns
    -------
    list[str]
        A list of function parameter python types, i.e. ['str', 'bool']
    """
    stringified_function_parameters: list[str] = []
    inputs_or_outputs = function.get(parameters_type, [])
    inputs_or_outputs = cast(list[ABIFunctionParams], inputs_or_outputs)

    for param in inputs_or_outputs:
        python_type = get_param_type(param)
        stringified_function_parameters.append(f"{python_type}")
    return stringified_function_parameters


def get_param_type(param: ABIFunctionParams | ABIFunctionComponents):
    """Gets the associated python type, including generated dataclasses"""
    internal_type = cast(str, param.get("internalType", ""))
    # if we find a struct, we'll add it to the dict of StructInfo's
    if is_struct(internal_type):
        python_type = get_struct_type(param)
    else:
        python_type = solidity_to_python_type(param.get("type", "unknown"))
    return python_type


def get_abi_from_json(json_abi: FoundryJson | HardhatJson | ABI) -> ABI:
    """Gets the ABI from a supported json format.
    Note that solc is not included here because it can have multiple contracts per json."""
    if is_foundry_json(json_abi):
        return _get_abi_from_foundry_json(json_abi)
    if is_hardhat_json(json_abi):
        return _get_abi_from_hardhat_json(json_abi)
    if is_abi(json_abi):
        return json_abi

    raise ValueError("Unable to identify an ABI for the given JSON.")


def is_abi(maybe_abi: object) -> TypeGuard[ABI]:
    """Typeguard for ABI's"""
    if not isinstance(json, list):
        return False  # ABI should be a list

    # Check if there's at least one entry with 'name', 'inputs', and 'type'
    for entry in maybe_abi:  # type: ignore
        if isinstance(entry, dict) and {"name", "inputs", "type"}.issubset(entry.keys()):
            return True

    return False  # No entry with the required fields was found


def _get_abi_from_foundry_json(json_abi: FoundryJson) -> ABI:
    return json_abi.get("abi")


def _get_abi_from_hardhat_json(json_abi: HardhatJson) -> ABI:
    return json_abi.get("abi")


def _get_abis_from_solc_json(json_abi: SolcJson) -> list[AbiInfo]:
    contracts = json_abi.get("contracts")
    infos: list[AbiInfo] = []

    for key, value in contracts.items():
        contract_name = key.split(":")[1]
        abi = value.get("abi")
        binary = value.get("bin")
        bytecode = f"0x{binary}"
        # TODO get bytecode link references and add here
        infos.append(AbiInfo(abi=abi, contract_name=contract_name, bytecode=bytecode, bytecode_link_references=[]))

    return infos



"""Functions to render Python types from an abi usng a jinja2 template."""

from __future__ import annotations

from pypechain.render.contract import ContractInfo
from pypechain.utilities.abi import get_structs_for_abi
from pypechain.utilities.templates import get_jinja_env


def render_types_file(contract_info: ContractInfo) -> str | None:
    """Returns the serialized code of the types file to be generated.

    Parameters
    ----------
    contract_info: ContractInfo
        Information related to the solidity contract.

    Returns
    -------
    str | None
        A serialized python file.
    """
    env = get_jinja_env()
    types_template = env.get_template("types.py.jinja2")

    structs = contract_info.structs.values()
    events = contract_info.events.values()
    errors = contract_info.errors.values()
    has_events = bool(events)
    has_structs = bool(structs)
    has_event_params = any(len(event.inputs) > 0 for event in events)

    structs_used = get_structs_for_abi(contract_info.abi)
    types_files_imported = {
        struct.contract_name for struct in structs_used if struct.contract_name != contract_info.contract_name
    }

    if not has_events and not has_structs:
        return None

    return types_template.render(
        contract_name=contract_info.contract_name,
        structs=structs,
        types_files_imported=list(types_files_imported),
        events=events,
        errors=errors,
        has_events=has_events,
        has_event_params=has_event_params,
    )



"""Functions to render Python files from an abi usng a jinja2 template."""

from __future__ import annotations

import os
from functools import partial
from multiprocessing import Pool
from pathlib import Path

from pypechain.render.contract import ContractInfo, get_contract_infos, render_contract_file
from pypechain.render.types import render_types_file
from pypechain.utilities.abi import AbiInfo
from pypechain.utilities.file import write_string_to_file
from pypechain.utilities.format import format_file


def render_files(
    abi_infos: list[AbiInfo],
    output_dir: str,
    line_length: int = 120,
    apply_formatting: bool = True,
    parallel: bool = False,
    chunksize: int | None = None,
) -> list[str]:
    """Processes a single JSON file to generate class and types files.


    Parameters
    ----------
    abi_infos : list[AbiInfo]
        A list of objects containing each ABI and metadata.
    output_dir : str
        The directory where files will be written to.
    line_length : int, optional
        Black's line-length config option.
    apply_formatting : bool, optional
        If True, autoflake, isort and black will be applied to the file in that order, by default True.
    parallel : bool, optional
        If True, multiple processes will be used to process the ABIs in parallel.
    chunksize: int, optional
        If set, abis are processed in approximately groups of chunksize.
        Use 1 for one process per ABI.
        Defaults to the number of ABIs // 10.

    Returns
    -------
    list[str]
        A list of filenames for the generated files.
    """
    # pylint: disable=too-many-arguments

    contract_infos = get_contract_infos(abi_infos)

    # this is what we are returning
    file_names: list[str] = []

    # for every [ContractName] generate a:
    #    1. [ContractName]Contract.py
    #    2. [ContractName]Types.py
    #  The Contract file defines the Contract class and functions.
    #  The Types file has the structs and events defined in the solidity contract.

    if parallel:
        # make a parallel pool; defaults to the number of available CPUs
        with Pool() as pool:
            # chunksize should be increased with the number of iterables
            chunksize = len(contract_infos) // 10
            for file_name_sublist in pool.imap(
                partial(
                    get_file_names, output_dir=output_dir, line_length=line_length, apply_formatting=apply_formatting
                ),
                list(contract_infos.values()),
                chunksize=chunksize,
            ):
                file_names.extend(file_name_sublist)
    # Sequential mode
    else:
        for contract_info in contract_infos.values():
            file_names.extend(
                get_file_names(
                    contract_info=contract_info,
                    output_dir=output_dir,
                    line_length=line_length,
                    apply_formatting=apply_formatting,
                )
            )

    return file_names


def get_file_names(contract_info: ContractInfo, output_dir: str, line_length: int, apply_formatting: bool) -> list[str]:
    """Get the file name for a single ContractInfo object


    Parameters
    ----------
    contract_info : ContractInfo
        A ContractInfo object created from an ABI.
    output_dir : str
        The directory where files will be written to.
    line_length : int, optional
        Black's line-length config option.
    apply_formatting : bool, optional
        If True, autoflake, isort and black will be applied to the file in that order, by default True.

    Returns
    -------
    list[str]
        A list of filenames for the generated Contract and Types files.
    """
    file_names: list[str] = []
    file_path = Path(output_dir)

    rendered_contract_code = render_contract_file(contract_info)
    if rendered_contract_code:
        contract_file_name = contract_info.contract_name + "Contract.py"
        contract_file_path = Path(os.path.join(file_path, contract_file_name))
        write_string_to_file(contract_file_path, rendered_contract_code)
        if apply_formatting is True:
            format_file(contract_file_path, line_length)
        file_names.append(f"{contract_info.contract_name}Contract")

    rendered_types_code = render_types_file(contract_info)
    if rendered_types_code:
        types_file_name = contract_info.contract_name + "Types.py"
        types_file_path = Path(os.path.join(file_path, types_file_name))
        write_string_to_file(types_file_path, rendered_types_code)
        if apply_formatting is True:
            format_file(types_file_path, line_length)
        file_names.append(f"{contract_info.contract_name}Types")

    return file_names



"""A web3.py Contract class for the OverloadedMethods contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain"""

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# This file is bound to get very long depending on contract sizes.
# pylint: disable=too-many-lines

# methods are overriden with specific arguments instead of generic *args, **kwargs
# pylint: disable=arguments-differ

# consumers have too many opinions on line length
# pylint: disable=line-too-long


from __future__ import annotations

from typing import Any, NamedTuple, Type, cast, overload

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from typing_extensions import Self
from web3 import Web3
from web3.contract.contract import Contract, ContractConstructor, ContractFunction, ContractFunctions
from web3.exceptions import FallbackNotFound
from web3.types import ABI, BlockIdentifier, CallOverride, TxParams

from .utilities import dataclass_to_tuple, rename_returned_types

structs = {}


class OverloadedMethodsDoSomethingContractFunction0(ContractFunction):
    """ContractFunction for the doSomething method."""

    def __call__(self, x: int, s: str) -> OverloadedMethodsDoSomethingContractFunction:  # type: ignore
        super().__call__(dataclass_to_tuple(x), dataclass_to_tuple(s))  # type: ignore
        return cast(OverloadedMethodsDoSomethingContractFunction, self)

    class ReturnValues(NamedTuple):
        """The return named tuple for DoSomething."""

        int_input: int
        arg2: str

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> ReturnValues:
        """returns ReturnValues."""
        # Define the expected return types from the smart contract call

        return_types = [int, str]

        # Call the function
        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)

        return self.ReturnValues(*rename_returned_types(structs, return_types, raw_values))


class OverloadedMethodsDoSomethingContractFunction1(ContractFunction):
    """ContractFunction for the doSomething method."""

    def __call__(self, s: str) -> OverloadedMethodsDoSomethingContractFunction:  # type: ignore
        super().__call__()  # type: ignore
        return cast(OverloadedMethodsDoSomethingContractFunction, self)

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> str:
        """returns str."""
        # Define the expected return types from the smart contract call

        return_types = str

        # Call the function
        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)

        return cast(str, rename_returned_types(structs, return_types, raw_values))


class OverloadedMethodsDoSomethingContractFunction2(ContractFunction):
    """ContractFunction for the doSomething method."""

    def __call__(self, x: int) -> OverloadedMethodsDoSomethingContractFunction:  # type: ignore
        super().__call__()  # type: ignore
        return cast(OverloadedMethodsDoSomethingContractFunction, self)

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> int:
        """returns int."""
        # Define the expected return types from the smart contract call

        return_types = int

        # Call the function
        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)

        return cast(int, rename_returned_types(structs, return_types, raw_values))


class OverloadedMethodsDoSomethingContractFunction3(ContractFunction):
    """ContractFunction for the doSomething method."""

    def __call__(self, x: int, y: int) -> OverloadedMethodsDoSomethingContractFunction:  # type: ignore
        super().__call__(dataclass_to_tuple(x), dataclass_to_tuple(y))  # type: ignore
        return cast(OverloadedMethodsDoSomethingContractFunction, self)

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> int:
        """returns int."""
        # Define the expected return types from the smart contract call

        return_types = int

        # Call the function
        raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)

        return cast(int, rename_returned_types(structs, return_types, raw_values))


class OverloadedMethodsDoSomethingContractFunction(ContractFunction):
    """ContractFunction for the doSomething method."""

    # super() call methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ# disable this warning when there is overloading
    # pylint: disable=function-redefined

    @overload
    def __call__(self, x: int, s: str) -> OverloadedMethodsDoSomethingContractFunction0:  # type: ignore
        ...

    @overload
    def __call__(self, s: str) -> OverloadedMethodsDoSomethingContractFunction1:  # type: ignore
        ...

    @overload
    def __call__(self, x: int) -> OverloadedMethodsDoSomethingContractFunction2:  # type: ignore
        ...

    @overload
    def __call__(self, x: int, y: int) -> OverloadedMethodsDoSomethingContractFunction3:  # type: ignore
        ...

    def __call__(self, *args) -> OverloadedMethodsDoSomethingContractFunction:  # type: ignore
        clone = super().__call__(*(dataclass_to_tuple(arg) for arg in args))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self  # type: ignore


class OverloadedMethodsContractFunctions(ContractFunctions):
    """ContractFunctions for the OverloadedMethods contract."""

    doSomething: OverloadedMethodsDoSomethingContractFunction

    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
        decode_tuples: bool | None = False,
    ) -> None:
        super().__init__(abi, w3, address, decode_tuples)
        self.doSomething = OverloadedMethodsDoSomethingContractFunction.factory(
            "doSomething",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="doSomething",
        )


overloadedmethods_abi: ABI = cast(
    ABI,
    [
        {
            "type": "function",
            "name": "doSomething",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "s", "type": "string", "internalType": "string"},
            ],
            "outputs": [
                {"name": "int_input", "type": "uint256", "internalType": "uint256"},
                {"name": "", "type": "string", "internalType": "string"},
            ],
            "stateMutability": "pure",
        },
        {
            "type": "function",
            "name": "doSomething",
            "inputs": [{"name": "s", "type": "string", "internalType": "string"}],
            "outputs": [{"name": "", "type": "string", "internalType": "string"}],
            "stateMutability": "pure",
        },
        {
            "type": "function",
            "name": "doSomething",
            "inputs": [{"name": "x", "type": "uint256", "internalType": "uint256"}],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "pure",
        },
        {
            "type": "function",
            "name": "doSomething",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [{"name": "added", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "pure",
        },
    ],
)


class OverloadedMethodsContract(Contract):
    """A web3.py Contract class for the OverloadedMethods contract."""

    abi: ABI = overloadedmethods_abi
    # We change `bytecode` as needed for linking, but keep
    # `_raw_bytecode` unchanged as an original copy.
    # pylint: disable=line-too-long
    _raw_bytecode: HexStr | None = HexStr(
        "0x608060405234801561001057600080fd5b5061033c806100206000396000f3fe608060405234801561001057600080fd5b506004361061004c5760003560e01c80633b6c27c4146100515780638ae3048e1461007a578063a6b206bf14610098578063b2dd1d79146100b9575b600080fd5b61006361005f366004610195565b9091565b604051610071929190610222565b60405180910390f35b61008b610088366004610243565b90565b6040516100719190610278565b6100ab6100a636600461028b565b6100cc565b604051908152602001610071565b6100ab6100c73660046102a4565b6100df565b60006100d98260026102dc565b92915050565b60006100eb82846102f3565b9392505050565b634e487b7160e01b600052604160045260246000fd5b600082601f83011261011957600080fd5b813567ffffffffffffffff80821115610134576101346100f2565b604051601f8301601f19908116603f0116810190828211818310171561015c5761015c6100f2565b8160405283815286602085880101111561017557600080fd5b836020870160208301376000602085830101528094505050505092915050565b600080604083850312156101a857600080fd5b82359150602083013567ffffffffffffffff8111156101c657600080fd5b6101d285828601610108565b9150509250929050565b6000815180845260005b81811015610202576020818501810151868301820152016101e6565b506000602082860101526020601f19601f83011685010191505092915050565b82815260406020820152600061023b60408301846101dc565b949350505050565b60006020828403121561025557600080fd5b813567ffffffffffffffff81111561026c57600080fd5b61023b84828501610108565b6020815260006100eb60208301846101dc565b60006020828403121561029d57600080fd5b5035919050565b600080604083850312156102b757600080fd5b50508035926020909101359150565b634e487b7160e01b600052601160045260246000fd5b80820281158282048414176100d9576100d96102c6565b808201808211156100d9576100d96102c656fea2646970667358221220b2ac531af898c14ca1c2bf29695116e40807da6cfffc0292534640d2d6903f4664736f6c63430008160033"
    )

    def __init__(self, address: ChecksumAddress | None = None) -> None:
        try:
            # Initialize parent Contract class
            super().__init__(address=address)
            self.functions = OverloadedMethodsContractFunctions(overloadedmethods_abi, self.w3, address)  # type: ignore

        except FallbackNotFound:
            print("Fallback function not found. Continuing...")

    functions: OverloadedMethodsContractFunctions

    @classmethod
    def constructor(cls) -> ContractConstructor:  # type: ignore
        """Creates a transaction with the contract's constructor function.

        Parameters
        ----------

        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.

        """
        cls.bytecode = cls._raw_bytecode
        if cls.bytecode is not None:

            # bytecode needs to be in hex for web3
            cls.bytecode = HexBytes(cls.bytecode)

        return super().constructor()

    @classmethod
    def deploy(cls, w3: Web3, account: LocalAccount | ChecksumAddress) -> Self:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        deployer = cls.factory(w3=w3)
        constructor_fn = deployer.constructor()

        # if an address is supplied, try to use a web3 default account
        if isinstance(account, str):
            tx_hash = constructor_fn.transact({"from": account})
            tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

            deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
            return deployed_contract

        # otherwise use the account provided.
        deployment_tx = constructor_fn.build_transaction()
        current_nonce = w3.eth.get_transaction_count(account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
        return deployed_contract

    @classmethod
    def factory(cls, w3: Web3, class_name: str | None = None, **kwargs: Any) -> Type[Self]:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        class_name: str | None
            The instance class name.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        contract = super().factory(w3, class_name, **kwargs)
        contract.functions = OverloadedMethodsContractFunctions(overloadedmethods_abi, w3, None)

        return contract



class OverloadedTransferContractEvent(ContractEvent):
    """ContractEvent for Transfer."""
    # super() get_logs and create_filter methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ

    # @combomethod destroys return types, so we are redefining functions as both class and instance
    # pylint: disable=function-redefined

    
    
    # pylint: disable=useless-parent-delegation
    def __init__(self, *argument_names: tuple[str]) -> None:
        super().__init__(*argument_names)

    def get_logs( # type: ignore
        self: "OverloadedTransferContractEvent",
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[EventData]:
        return cast(Iterable[EventData], super().get_logs(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, block_hash=block_hash))

    @classmethod
    def get_logs( # type: ignore
        cls: Type["OverloadedTransferContractEvent"],
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[EventData]:
        return cast(Iterable[EventData], super().get_logs(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, block_hash=block_hash))

    def create_filter( # type: ignore
        self: "OverloadedTransferContractEvent",
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(LogFilter, super().create_filter(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, address=address, topics=topics))

    @classmethod
    def create_filter( # type: ignore
        cls: Type["OverloadedTransferContractEvent"],
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(LogFilter, super().create_filter(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, address=address, topics=topics))

class OverloadedApprovalContractEvent(ContractEvent):
    """ContractEvent for Approval."""
    # super() get_logs and create_filter methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ

    # @combomethod destroys return types, so we are redefining functions as both class and instance
    # pylint: disable=function-redefined

    
    
    # pylint: disable=useless-parent-delegation
    def __init__(self, *argument_names: tuple[str]) -> None:
        super().__init__(*argument_names)

    def get_logs( # type: ignore
        self: "OverloadedApprovalContractEvent",
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[EventData]:
        return cast(Iterable[EventData], super().get_logs(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, block_hash=block_hash))

    @classmethod
    def get_logs( # type: ignore
        cls: Type["OverloadedApprovalContractEvent"],
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[EventData]:
        return cast(Iterable[EventData], super().get_logs(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, block_hash=block_hash))

    def create_filter( # type: ignore
        self: "OverloadedApprovalContractEvent",
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(LogFilter, super().create_filter(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, address=address, topics=topics))

    @classmethod
    def create_filter( # type: ignore
        cls: Type["OverloadedApprovalContractEvent"],
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        fromBlock: BlockIdentifier | None = None,
        toBlock: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(LogFilter, super().create_filter(argument_filters=argument_filters, fromBlock=fromBlock, toBlock=toBlock, address=address, topics=topics))



class OverloadedContractEvents(ContractEvents):
    """ContractEvents for the Overloaded contract."""

    Transfer: OverloadedTransferContractEvent

    Approval: OverloadedApprovalContractEvent


    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
    ) -> None:
        super().__init__(abi, w3, address)
        self.Transfer = cast(OverloadedTransferContractEvent, OverloadedTransferContractEvent.factory(
            "Transfer",
            w3=w3,
            contract_abi=abi,
            address=address,
            event_name="Transfer"
        ))
        self.Approval = cast(OverloadedApprovalContractEvent, OverloadedApprovalContractEvent.factory(
            "Approval",
            w3=w3,
            contract_abi=abi,
            address=address,
            event_name="Approval"
        ))
        


"""Dataclasses for all structs in the IStructs contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name
# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import
# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return
from __future__ import annotations

from dataclasses import dataclass


@dataclass
class InnerStruct:
    """InnerStruct struct."""

    boolVal: bool


@dataclass
class SimpleStruct:
    """SimpleStruct struct."""

    intVal: int
    strVal: str


@dataclass
class NestedStruct:
    """NestedStruct struct."""

    intVal: int
    strVal: str
    innerStruct: InnerStruct


@dataclass
class ErrorInfo:
    """Custom contract error information."""

    name: str
    selector: str
    signature: str
    inputs: list[ErrorParams]


@dataclass
class ErrorParams:
    """Parameter info for custom contract errors."""

    name: str
    solidity_type: str
    python_type: str
